<?php

declare(strict_types=1);

/**
 * @author Amasty Team
 * @copyright Copyright (c) Amasty (https://www.amasty.com)
 * @package Omnibus Price Tracker by Amasty
 */

namespace Amasty\PriceHistory\Model\Frontend\DisplayFilter;

use Amasty\PriceHistory\Model\ConfigProvider;
use Amasty\PriceHistory\Model\OptionSource\DisplayMode;
use Magento\Bundle\Model\Product\Type as Bundle;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Pricing\Price\FinalPrice;
use Magento\Catalog\Pricing\Price\RegularPrice;
use Magento\Catalog\Pricing\Price\SpecialPrice;
use Magento\ConfigurableProduct\Model\Product\Type\Configurable;
use Magento\Framework\Stdlib\DateTime\TimezoneInterface;

class OnSaleFilter implements FilterInterface
{
    /**
     * @var ConfigProvider
     */
    private ConfigProvider $configProvider;

    /**
     * @var TimezoneInterface
     */
    private TimezoneInterface $timezone;

    /**
     * @var array
     */
    private array $skipableProductTypes;

    public function __construct(
        ConfigProvider $configProvider,
        TimezoneInterface $timezone,
        array $skipableProductTypes = []
    ) {
        $this->configProvider = $configProvider;
        $this->timezone = $timezone;
        $this->skipableProductTypes = $skipableProductTypes;
    }

    public function show(ProductInterface $product): bool
    {
        $storeId = (int)$product->getStoreId();
        if ($this->configProvider->getDisplayMode($storeId) === DisplayMode::DISCOUNTED) {
            return $this->isProductDiscounted($product);
        }

        return true;
    }

    private function isProductDiscounted(ProductInterface $product): bool
    {
        if (in_array($product->getTypeId(), $this->skipableProductTypes)) {
            return false;
        }

        $specialPrice = $this->getSpecialPrice($product);
        $regularPrice = $this->getPriceByCode($product, RegularPrice::PRICE_CODE);

        return $specialPrice < $regularPrice;
    }

    private function getSpecialPrice(ProductInterface $product): float
    {
        switch ($product->getTypeId()) {
            case Bundle::TYPE_CODE:
            case Configurable::TYPE_CODE:
                $specialPrice = $this->getPriceByCode($product, FinalPrice::PRICE_CODE);
                break;
            default:
                if ($this->isSpecialPriceActive($product)) {
                    $specialPrice = $this->getPriceByCode($product, SpecialPrice::PRICE_CODE);
                    if (!$specialPrice) {
                        $specialPrice = $this->getPriceByCode($product, RegularPrice::PRICE_CODE);
                    }
                } else {
                    $specialPrice = $this->getPriceByCode($product, FinalPrice::PRICE_CODE);
                }
        }

        return (float)$specialPrice;
    }

    /**
     * @param ProductInterface $product
     * @param string $code
     *
     * @return float|string|false
     */
    private function getPriceByCode(ProductInterface $product, string $code)
    {
        return $product->getPriceInfo()->getPrice($code)->getAmount()->getValue();
    }

    private function isSpecialPriceActive(ProductInterface $product): bool
    {
        $specialFromDate = $product->getSpecialFromDate();
        $specialToDate = $product->getSpecialToDate();

        return ($specialFromDate || $specialToDate)
            && $this->timezone->isScopeDateInInterval(
                $product->getStoreId(),
                $specialFromDate,
                $specialToDate
            );
    }
}
