<?php

declare(strict_types=1);

/**
 * @author Amasty Team
 * @copyright Copyright (c) Amasty (https://www.amasty.com)
 * @package Omnibus Price Tracker by Amasty
 */

namespace Amasty\PriceHistory\Model\PriceHistory\ResourceModel;

use Amasty\PriceHistory\Model\PriceHistory\PriceHistory as PriceHistoryModel;
use Magento\Framework\DB\Select;
use Magento\Framework\Model\ResourceModel\Db\AbstractDb;

class PriceHistory extends AbstractDb
{
    public const TABLE_NAME = 'amasty_price_history';

    protected function _construct()
    {
        $this->_init(self::TABLE_NAME, PriceHistoryModel::ID);
    }

    public function getMinPriceExpr(): \Zend_Db_Expr
    {
        return new \Zend_Db_Expr('IF (final_price != 0 AND final_price < min_price, final_price, min_price)');
    }

    public function getProductLowestPrice(int $productId, string $datetime, int $websiteId, int $customerGroupId): float
    {
        $connection = $this->getConnection();

        $conditions = [
            PriceHistoryModel::PRODUCT_ID . ' = ?' => $productId,
            PriceHistoryModel::WEBSITE_ID . ' = ?' => $websiteId,
            PriceHistoryModel::CUSTOMER_GROUP_ID . ' = ?' => $customerGroupId,
            PriceHistoryModel::DATE . ' > ?' => $datetime
        ];
        $select = $this->prepareLastPricesSelect($conditions);

        return $this->getLowestPrice($connection->fetchCol($select));
    }

    public function clean(?int $period = null): void
    {
        $connection = $this->getConnection();
        $tableName = $this->getMainTable();

        if ($period === null) {
            $connection->delete($tableName);
        } else {
            $connection->delete(
                $tableName,
                PriceHistoryModel::DATE . ' < (CURRENT_TIMESTAMP - INTERVAL ' . $period . ' DAY)'
            );
        }
    }

    public function cleanByWebsiteIds(array $websiteIds = []): void
    {
        $connection = $this->getConnection();
        $tableName = $this->getMainTable();

        $connection->delete(
            $tableName,
            [PriceHistoryModel::WEBSITE_ID . ' IN (?)' => $websiteIds]
        );
    }

    private function prepareLastPricesSelect(array $conditions): Select
    {
        $connection = $this->getConnection();

        $select = $connection->select()->from(
            $this->getMainTable(),
            [PriceHistoryModel::NEW_PRICE]
        );
        foreach ($conditions as $cond => $value) {
            $select->where($cond, $value);
        }
        $select->order(PriceHistoryModel::DATE . ' DESC');

        return $select;
    }

    /**
     * @param string[] $prices
     */
    private function getLowestPrice(array $prices): float
    {
        $price = 0;
        if (empty($prices)) {
            return (float)$price;
        }

        switch (count($prices)) {
            case 1:
                $price = $prices[0];
                break;
            case 2:
                $price = $prices[1];
                break;
            default:
                $prices = array_slice($prices, 1);
                $price = min($prices);
                break;
        }

        return (float)$price;
    }
}
