<?php

declare(strict_types=1);

/**
 * @author Amasty Team
 * @copyright Copyright (c) Amasty (https://www.amasty.com)
 * @package Omnibus Price Tracker by Amasty
 */

namespace Amasty\PriceHistory\Test\Integration\Model\PriceHistory\Processor;

use Amasty\PriceHistory\Model\PriceHistory\PriceHistory;
use Amasty\PriceHistory\Model\PriceHistory\ResourceModel\Collection;
use Amasty\PriceHistory\Model\PriceHistory\ResourceModel\IndexProcessorActions\FillLatestPriceTable;
use Amasty\PriceHistory\Model\PriceHistory\ResourceModel\PriceHistory as PriceHistoryResource;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\ResourceModel\Product as ProductResource;
use Magento\Framework\Indexer\IndexerInterface;
use Magento\Framework\Indexer\IndexerRegistry;
use Magento\Framework\ObjectManagerInterface;
use Magento\TestFramework\Helper\Bootstrap;
use PHPUnit\Framework\TestCase;

/**
 * @magentoAppIsolation enabled
 * @magentoDbIsolation disabled
 */
class IndexProcessorTest extends TestCase
{
    /**
     * @var ObjectManagerInterface
     */
    private $objectManager;

    /**
     * @var IndexerInterface
     */
    private $priceIndexer;

    /**
     * @var ProductRepositoryInterface
     */
    private $productRepository;

    protected function setUp(): void
    {
        $this->objectManager = Bootstrap::getObjectManager();

        $indexRegistry = $this->objectManager->get(IndexerRegistry::class);
        $this->priceIndexer = $indexRegistry->get('catalog_product_price');
        $this->priceIndexer->setScheduled(true);

        $this->productRepository = $this->objectManager->get(ProductRepositoryInterface::class);
    }

    /**
     * @magentoDataFixture Amasty_PriceHistory::Test/Integration/_files/fill_price_history_table.php
     * @magentoDataFixture Amasty_PriceHistory::Test/Integration/_files/catalog_product_price_reindex_all.php
     *
     * @magentoConfigFixture current_store price_history/general/enable 1
     */
    public function testNewRecords(): void
    {
        $product = $this->productRepository->get('simple');
        $newPrice = $product->getPrice() - 5;
        $this->assertProductPriceUpdateLogged($product, (float)$newPrice);
    }

    protected function tearDown(): void
    {
        $this->clearTable('catalog_product_index_price');
        $this->clearTable(PriceHistoryResource::TABLE_NAME);
        $this->clearTable(FillLatestPriceTable::LATEST_PRICE_TABLE_NAME);
    }

    /**
     * @magentoDataFixture Magento/Catalog/_files/product_simple.php
     * @magentoDataFixture Amasty_PriceHistory::Test/Integration/_files/catalog_product_price_reindex_all.php
     *
     * @magentoConfigFixture current_store price_history/general/enable 1
     */
    public function testNoNewRecords(): void
    {
        $priceHistoryCollection = $this->objectManager->create(Collection::class);
        $lastRecordBefore = $priceHistoryCollection->getLastItem();

        $indexRegistry = $this->objectManager->get(IndexerRegistry::class);
        $priceIndexer = $indexRegistry->get('catalog_product_price');
        $priceIndexer->reindexAll();

        $priceHistoryCollection = $this->objectManager->create(Collection::class);
        $lastRecordAfter = $priceHistoryCollection->getLastItem();

        $this->assertEquals($lastRecordBefore->getId(), $lastRecordAfter->getId());
    }

    /**
     * @magentoDataFixture Magento/Catalog/_files/product_simple.php
     * @magentoDataFixture Amasty_PriceHistory::Test/Integration/_files/catalog_product_price_reindex_all.php
     *
     * @magentoConfigFixture current_store price_history/general/enable 1
     */
    public function testAfterClear(): void
    {
        $this->clearTable(PriceHistoryResource::TABLE_NAME);

        $product = $this->productRepository->get('simple');
        $newPrice = $product->getPrice() + 5;
        $this->assertProductPriceUpdateLogged($product, (float)$newPrice);
    }

    /**
     * @magentoDataFixture Magento/Bundle/_files/bundle_product_with_dynamic_price.php
     * @magentoDataFixture Amasty_PriceHistory::Test/Integration/_files/catalog_product_price_reindex_all.php
     *
     * @magentoConfigFixture current_store price_history/general/enable 1
     */
    public function testBundleProduct(): void
    {
        $product = $this->productRepository->get('bundle_product_with_dynamic_price');
        $oldPrice = $product->getPriceInfo()->getPrice('final_price')->getAmount()->getValue();
        $reindexIds[] = $product->getId();

        $childrens = $product->getTypeInstance()->getChildrenIds($product->getId(), true);
        foreach ($childrens as $childProductLink) {
            $childProductId = $childProductLink[array_key_last($childProductLink)];
            $childProduct = $this->productRepository->getById($childProductId);
            $reindexIds[] = $childProduct->getId();
            $this->updateProductPrice($childProduct, 5);
        }
        $this->priceIndexer->reindexList($reindexIds);
        $priceHistoryCollection = $this->objectManager->create(Collection::class)
            ->addFieldToFilter(PriceHistory::PRODUCT_ID, $product->getId());
        $lastRecord = $priceHistoryCollection->getLastItem();

        $this->assertEquals(count($childrens) * 5, $lastRecord->getNewPrice());
        $this->assertEquals($oldPrice, $lastRecord->getLatestPrice());
    }

    private function assertProductPriceUpdateLogged(Product $product, float $newPrice): void
    {
        $previousPrice = $product->getPrice();
        $this->updateProductPrice($product, $newPrice);

        $this->priceIndexer->reindexRow($product->getId());
        $priceHistoryCollection = $this->objectManager->create(Collection::class);
        $lastRecord = $priceHistoryCollection->getLastItem();

        $this->assertEquals($product->getId(), $lastRecord->getProductId()); //covers logging of right product
        $this->assertEquals($newPrice, $lastRecord->getNewPrice());
        $this->assertEquals($previousPrice, $lastRecord->getLatestPrice()); //covers AddLatestPrice
        $this->assertNotEquals($lastRecord->getNewPrice(), $lastRecord->getLatestPrice()); //covers FilterUnique
    }

    /**
     * Must use direct query because saving through model or repo
     * blocks indexer for repeating in current transaction
     */
    private function updateProductPrice(Product $product, float $value): void
    {
        $priceAttrId = $product->getAttributes()['price']->getId();
        $resource = $this->objectManager->get(ProductResource::class);
        $linkField = $resource->getLinkField();
        $connection = $resource->getConnection();
        $connection->update(
            $resource->getTable('catalog_product_entity_decimal'),
            ['value' => $value],
            ['attribute_id = ?' => $priceAttrId, $linkField . ' = ?' => $product->getData($linkField)]
        );
    }

    private function clearTable(string $tableName): void
    {
        $resource = $this->objectManager->get(ProductResource::class);
        $connection = $resource->getConnection();
        $tableName = $resource->getTable($tableName);

        $connection->delete($tableName);
    }
}
